'''
This is not part of specification

Helper class to make it work as python lib
'''

from .base import Decorator, ATTR_KEY


class DbusAttr(Decorator):
    '''
    https://docs.python.org/2/howto/descriptor.html#properties
    '''

    def __init__(self, meth=None, produces=lambda resp: resp):
        self.attr = meth
        self.produces = produces
        self._update_me(meth)

    def __call__(self, meth):
        self.attr = meth
        self._update_me(meth)
        return self
    
    def __get__(self, obj, objtype=None):
        #static call
        if not obj:
            return self

        _dbus = getattr(obj, ATTR_KEY)
        props = _dbus.properties
        iface = _dbus.iface
        result = props.Get(iface, self.attr.__name__)
        produces = self.produces
        return produces(result)

    def __set__(self, obj, value):
        if obj:
            _dbus = getattr(obj, ATTR_KEY)
            props = _dbus.properties
            iface = _dbus.iface
            props.Set(iface, self.attr.__name__, value)
        else: #static call
            self.attr = value

    def __delete__(self, obj):
        raise AttributeError('can not delete attribute')


if __name__ == '__main__':
    # examples
    from .interface import DbusInterface

    @DbusInterface('org.mpris.MediaPlayer2',
                   '/org/mpris/MediaPlayer2')
    class Example(object):
        @DbusAttr
        def Identity(self):
            pass

    d = Example(
        dbus_interface_info={
            'dbus_uri': 'org.mpris.MediaPlayer2.vlc'})

    assert d.Identity == 'VLC media player'



'''
From mprisV2.2 documentation

http://specifications.freedesktop.org/mpris-spec/latest/Playlists_Interface.html
'''

from .decorator import DbusAttr
from .decorator import DbusInterface
from .decorator import DbusMethod
from .decorator import DbusSignal
from .interfaces import Interfaces
from .types import Playlist, Maybe_Playlist
from dbus import UInt32


class Playlists(Interfaces):
    '''
    Provides access to the media player's playlists.
    
    Since D-Bus does not provide an easy way to check for what interfaces are exported on an object, clients should attempt to get one of the properties on this interface to see if it is implemented.
    
    '''
    
    @DbusInterface(Interfaces.PLAYLISTS, Interfaces.OBJECT_PATH)
    def __init__(self):
        '''Constructor'''
        pass
    
    @DbusMethod
    def ActivatePlaylist(self, PlaylistId):
        '''
        **Parameters:**
        
        * PlaylistId - o
            The id of the playlist to activate.

        Starts playing the given playlist.
        
        Note that this must be implemented. If the media player does not allow clients to change the playlist, it should not implement this interface at all.
            
        It is up to the media player whether this completely replaces the current tracklist, or whether it is merely inserted into the tracklist and the first track starts. For example, if the media player is operating in a 'jukebox' mode, it may just append the playlist to the list of upcoming tracks, and skip to the first track in the playlist.
        '''
        pass
    
    @DbusMethod(produces=lambda playlist_list: \
                    [Playlist(playlist) for playlist in playlist_list],
                args_to_dbus=[UInt32, UInt32, str, bool])
    def GetPlaylists(self, Index, MaxCount, Order, ReverseOrder=False):
        '''
        **Parameters:**
        
        * Index - u
            The index of the first playlist to be fetched (according to the ordering).
        * MaxCount - u
            The maximum number of playlists to fetch.
        * Order - s (Playlist_Ordering)
            The ordering that should be used.
        * ReverseOrder - b
            Whether the order should be reversed.

        **Returns**
        
        * Playlists - a(oss) (Playlist_List)
            A list of (at most MaxCount) playlists.
    
        Gets a set of playlists.
        '''
        pass
    
    @DbusSignal
    def PlaylistChanged(self, Playlist):
        '''
        **Parameters**
        
        * Playlist - (oss) (Playlist)
            The playlist whose details have changed.
    
        Indicates that the name or icon for a playlist has changed.
        
        Note that, for this signal to operate correctly, the id of the playlist must not change when the name changes.
        '''
        pass
    
    @DbusAttr
    def PlaylistCount(self):
        '''
        **Returns**
        
        Read only
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.
    
        The number of playlists available.
        '''
        pass
    
    @DbusAttr
    def Orderings(self):
        '''
        **Returns**
        
        Read only
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.
    
        The avaislable orderings. At least one must be offered.
        '''
        pass
    
    @DbusAttr(produces=Maybe_Playlist)
    def ActivePlaylist(self):
        '''
        **Returns**
        
        Read only
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.
    
        The currently-active playlist.
        
        If there is no currently-active playlist, the structure's Valid field will be false, and the Playlist details are undefined.
        
        Note that this may not have a value even after ActivatePlaylist is called with a valid playlist id as ActivatePlaylist implementations have the option of simply inserting the contents of the playlist into the current tracklist.
        '''
        pass


if __name__ == '__main__':
    from .utils import get_players_uri, implements
    for uri in get_players_uri():
        if implements(uri, Interfaces.PLAYLISTS):
            mp2 = Playlists(dbus_interface_info={'dbus_uri': uri})
            print( mp2.ActivePlaylist )
            print( 'Active is valid playlist: ', bool(mp2.ActivePlaylist) )
            if mp2.ActivePlaylist:
                print( 'Active playlist name:', mp2.ActivePlaylist.Playlist.Name )
            from mpris2.types import Playlist_Ordering
            print( hasattr('anystring', 'eusequenaotem') )
            print( 'bla', mp2.GetPlaylists(0, 20,
                                           Playlist_Ordering.ALPHABETICAL, False) )
            break
    else:
        print('no player with playlist interface found')



class Some_Players(object):
    '''
    Not defined in documentation
    
    Maybe this player (and other) implement mpris2
    
     **Some players**
    
    * AUDACIOUS
        'audacious'
    * BANSHEE 
        'banshee'
    * BEATBOX
        'beatbox'
    * BMP
        'bmp'
    * CLEMENTINE
        'clementine'
    * DRAGONPLAYER
        'dragonplayer'
    * EXAILE
        'exaile'
    * GMUSICBROWSER
        'gmusicbrowser'
    * GMPC
        'gmpc'
    * GUAYADEQUE
        'guayadeque'
    * MOPIDY
        'mopidy'
    * MPDRIS
        'mpDris'
    * QUODLIBET
        'quodlibet'
    * RAVEND
        'ravend'
    * RHYTHMBOX
        'rhythmbox'
    * SPOTIFY
        'spotify'
    * VLC
        'vlc'
    * XBMC
        'xbmc'
    * XMMS2
        'xmms2'
    * XNOISE
        'xnoise'
    '''
    AUDACIOUS = 'audacious'
    BANSHEE = 'banshee'
    BEATBOX = 'beatbox'
    BMP = 'bmp'
    CLEMENTINE = 'clementine'
    DRAGONPLAYER = 'dragonplayer'
    EXAILE = 'exaile'
    GMUSICBROWSER = 'gmusicbrowser'
    GMPC = 'gmpc'
    GUAYADEQUE = 'guayadeque'
    MOPIDY = 'mopidy'
    MPDRIS = 'mpDris'
    QUODLIBET = 'quodlibet'
    RAVEND = 'ravend'
    RHYTHMBOX = 'rhythmbox'
    SPOTIFY = 'spotify'
    VLC = 'vlc'
    XBMC = 'xbmc'
    XMMS2 = 'xmms2'
    XNOISE = 'xnoise'
    
    @staticmethod
    def get_dict():
        result  = {}
        for key in dir(Some_Players):
            if key[0] not in ('_', 'g'):
                result[key] = getattr(Some_Players, key)
        return result


if __name__ == '__main__':
    print('Well know players:', Some_Players.get_dict().values())



'''
From mprisV2.2 documentation

http://specifications.freedesktop.org/mpris-spec/2.2/Playlists_Interface.html#Simple-Type:Uri
'''


class Uri(str):
    '''A unique resource identifier.'''

    def __init__(self, uri):
        self._uri = uri
    
    def __str__(self):
        return str(self._uri)
    
if __name__ == '__main__':
    assert Uri('http://www.com.br') == 'http://www.com.br'



'''
From mprisV2.2 documentation

http://specifications.freedesktop.org/mpris-spec/latest/Player_Interface.html#Simple-Type:Playback_Rate
'''


class Playback_Rate(float):
    '''
    A playback rate
    
    This is a multiplier,
    so a value of 0.5 indicates that playback is happening at half speed,
    while 1.5 means that 1.5 seconds of 'track time' is consumed every second.
    '''

    def __init__(self, rate=1.0):
        self._rate = rate
    
    def __float__(self):
        return self._rate
    
    def __str__(self):
        return str(self._rate)

    
if __name__ == '__main__':
    pr = Playback_Rate(1.2)
    assert pr == 1.2



'''
From mprisV2.2 documentation

http://specifications.freedesktop.org/mpris-spec/2.2/Playlists_Interface.html#Struct:Playlist
'''

from dbus import Struct


class Playlist(Struct):
    '''
    A data structure describing a playlist.
    
    * Id - o (Playlist_Id)
        A unique identifier for the playlist.
        
        This should remain the same if the playlist is renamed.
    * Name - s
        The name of the playlist, typically given by the user.
    * Icon - s (Uri)
        The URI of an (optional) icon.
        
    '''
    def __init__(self, playlist):
        Struct.__init__(
           self,
           iter(playlist),
           signature=playlist.signature,
           variant_level=playlist.variant_level
        )
    
    @property
    def Id(self):
        return self[0]
    
    @property
    def Name(self):
        return self[1]
    
    @property
    def Icon(self):
        return self[2]


class Maybe_Playlist(Struct):
    '''
    * Valid - b
        Whether this structure refers to a valid playlist.
    * Playlist - (oss) (Playlist)
        The playlist, providing Valid is true, otherwise undefined.
        
    When constructing this type, it should be noted that the playlist ID must 
    be a valid object path, or D-Bus implementations may reject it. This is
    true even when Valid is false. It is suggested that '/' is used as the
    playlist ID in this case.
    '''
    
    def __init__(self, maybe_playlist=None):
        Struct.__init__(
           self,
           (maybe_playlist[0], maybe_playlist[1]),
           signature=maybe_playlist.signature,
           variant_level=maybe_playlist.variant_level
        )

    @property
    def Valid(self):
        return self[0]
    
    @property
    def Playlist(self):
        return Playlist(self[1])
    
    def __nonzero__(self):
        return bool(self.Valid)
    
    def __bool__(self):
        return self.__nonzero__()


if __name__ == '__main__':
    expect = Struct((1, 'My Playlist', None), signature='iss')
    p = Playlist(expect)
    assert p.Id == 1
    assert p.Name == 'My Playlist'
    assert p.Icon == None
    m_expect = Struct((False, p))
    mp = Maybe_Playlist(m_expect)
    assert not mp.Valid
    assert not mp



#!/usr/bin/python
# -*- coding: UTF8 -*
'''
This is a copy of mprisV2.2 documentation

http://specifications.freedesktop.org/mpris-spec/latest/

That also works as python lib.



Version 2.2
===========
Copyright © 2006-2010 the VideoLAN team(Mirsal Ennaime, Rafaël Carré, Jean-Paul Saman)

Copyright © 2005-2008 Milosz Derezynski

Copyright © 2008 Nick Welch

Copyright © 2010-2012 Alex Merry

This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.



About
=====
The Media Player Remote Interfacing Specification is a standard D-Bus interface which aims to provide a common programmatic API for controlling media players.

It provides a mechanism for compliant media players discovery, basic playback and media player state control as well as a tracklist interface which is used to add context to the current item.



Changes
=======
 Changes (Permalink)

From 2.1 to 2.2:

* Added the optional Fullscreen and CanSetFullscreen properties to the org.mpris.MediaPlayer2 interface.
* The path /org/mpris/MediaPlayer2/TrackList/NoTrack now represents "no track" where required in the org.mpris.MediaPlayer2.TrackList interface (since empty paths are not allowed by D-Bus).
* The suggested unique instance identifier no longer violates the D-Bus specification by begining with a digit.


From 2.0 to 2.1:

* Added the optional org.mpris.MediaPlayer2.Playlists interface.



Bus Name Policy
===============
Each media player *must* request a unique bus name which begins with *org.mpris.MediaPlayer2*. For example:

* org.mpris.MediaPlayer2.audacious
* org.mpris.MediaPlayer2.vlc
* org.mpris.MediaPlayer2.bmp
* org.mpris.MediaPlayer2.xmms2

This allows clients to list available media players (either already running or which can be started via D-Bus activation)

In the case where the media player allows multiple instances running simultaneously, each additional instance should request a unique bus name, adding a dot and a unique identifier to its usual bus name, such as one based on a UNIX process id. For example, this could be:

* org.mpris.MediaPlayer2.vlc.7389

Note: According to the D-Bus specification, the unique identifier "must only contain the ASCII characters '[A-Z][a-z][0-9]_-'" and "must not begin with a digit". 

Entry point
===========
The media player *must* expose the */org/mpris/MediaPlayer2* object path, which *must* implement the following interfaces:

* :doc:`org.mpris.MediaPlayer2<../mpris2.mediaplayer2>`
* :doc:`org.mpris.MediaPlayer2.Player<../mpris2.player>`

The */org/mpris/MediaPlayer2* object may implement the :doc:`org.mpris.MediaPlayer2.TrackList<../mpris2.tracklist>` interface.

The */org/mpris/MediaPlayer2* object may implement the :doc:`org.mpris.MediaPlayer2.Playlists<../mpris2.playlists>` interface.


The PropertiesChanged signal
============================
The MPRIS uses the org.freedesktop.DBus.Properties.PropertiesChanged signal to notify clients of changes in the media player state. If a client implementation uses D-Bus bindings which do not support this signal, then it should connect to it manually. If a media player implementation uses D-Bus bindings which do not support this signal, then it should send it manually


Corrections
===========
2010-09-26: Added EmitsChangedSignal annotation to Volume property on the Player interface.

2011-01-26: Added PlaylistChanged signal to the Playlists interface.


Interfaces
==========
* :doc:`org.mpris.MediaPlayer2<../mpris2.mediaplayer2>`
* :doc:`org.mpris.MediaPlayer2.Player<../mpris2.player>`
* :doc:`org.mpris.MediaPlayer2.Playlists<../mpris2.playlists>`
* :doc:`org.mpris.MediaPlayer2.TrackList<../mpris2.tracklist>`


'''

from .interfaces import Interfaces
from .mediaplayer2 import MediaPlayer2
from .player import Player
from .playlists import Playlists
from .tracklist import TrackList
from .utils import get_players_uri
from .some_players import Some_Players as SomePlayers



'''
This is not part of specification

Helper class to make it work as python lib
'''

from .base import Decorator, ATTR_KEY


def kw_to_dbus(**kw):
    return kw

def args_to_dbus(*args):
    return args


class DbusMethod(Decorator):

    def __init__(self, meth=None,
                 iface=None,
                 produces=lambda resp: resp,
                 args_to_dbus=args_to_dbus,
                 kw_to_dbus=kw_to_dbus,
                 std_args=(),
                 std_kwds={}):
        self.meth = meth
        self.handler = None
        self.produces = produces
        self.iface = iface
        self.args_to_dbus = args_to_dbus
        self.kw_to_dbus = kw_to_dbus
        self.std_args = std_args
        self.std_kwds = std_kwds
        self.obj = None
        self._update_me(meth)

    def __call__(self, meth=None):
        self.meth = meth
        self._update_me(meth)
        return self

    def __get__(self, obj=None, cls=None):
        if obj is None:
            return self
        self.obj = obj
        return self._call_dbus
    
    def _call_dbus(self, *args, **kwds):
        _dbus = getattr(self.obj, ATTR_KEY)
        if self.iface:
            iface = self.iface
        else:
            iface = _dbus.iface
        bus_obj = _dbus.object
        bus_meth = bus_obj.get_dbus_method(self.meth.__name__, iface)
        _args = self.merge_args(args, self.std_args)
        args = self.convert_args_to_dbus_args(*_args)
        _kwds = self.std_kwds.copy()
        _kwds.update(kwds)
        kwds = self.convert_kw_to_dbus_kw(**_kwds)
        result = bus_meth(*args, **kwds)
        return self.produces(result)
    
    @classmethod
    def merge_args(cls, args, std_args):
        _len = len(std_args) - len(args)
        return args + std_args[-_len:] if _len > 0 else args
    
    @classmethod
    def merge_kwds(cls, kwds, std_kwds):
        _kwds = std_kwds.copy()
        _kwds.update(kwds)
        return _kwds
    
    def convert_args_to_dbus_args(self, *args):
        args_to_dbus = self.args_to_dbus
        if callable(args_to_dbus):
            return args_to_dbus(*args)
        
        #iterate over args
        result = []
        for arg in args:
            i = args.index(arg)
            if i < len(args_to_dbus):
                make = args_to_dbus[i]
                if callable(make):
                    arg = make(arg)
            result.append(arg)            
        return tuple(result)
    
    def convert_kw_to_dbus_kw(self, **kw):
        kw_to_dbus = self.kw_to_dbus
        if callable(kw_to_dbus):
            return kw_to_dbus(**kw)
        
        if hasattr(self.kw_to_dbus, 'keys'):
            for key, val in kw.items():
                make = kw_to_dbus.get(key, lambda v: v)
                kw[key] = make(val)
        return kw


if __name__ == '__main__':
    # examples
    from .interface import DbusInterface
    @DbusInterface('org.freedesktop.DBus', '/')
    class Example(object):
        
        @DbusMethod
        def GetId(self):
            pass
        
        @DbusMethod
        def GetNameOwner(self, name):
            pass

    d = Example(
        dbus_interface_info={
            'dbus_uri': 'org.freedesktop.DBus'})
    assert d.GetId()
    assert d.GetNameOwner('org.freedesktop.DBus') == 'org.freedesktop.DBus'



'''
From mprisV2.2 documentation

http://specifications.freedesktop.org/mpris-spec/latest/Player_Interface.html#Simple-Type:Time_In_Us
'''


class Time_In_Us(int):
    '''Time in microseconds.'''
    
    def __init__(self, time=0):
        self._time = time
    
    def __int__(self):
        return int(self._time)
    
    def __str__(self):
        return str(self._time)


if __name__ == '__main__':
    assert Time_In_Us(10) == 10



'''
From mprisV2.2 documentation

http://specifications.freedesktop.org/mpris-spec/2.2/#Interfaces
'''


class Interfaces(object):
    '''
    This class contains the constants defined at index of MPRIS2 definition:

    
    **Interfaces:**
    
    * MEDIA_PLAYER
        'org.mpris.MediaPlayer2'
    * TRACK_LIST
        'org.mpris.MediaPlayer2.TrackList'
    * PLAYER
        'org.mpris.MediaPlayer2.Player'
    * PLAYLISTS
        'org.mpris.MediaPlayer2.Playlists'
    * PROPERTIES
        'org.freedesktop.DBus.Properties'
    

    **Signals:**
    
    * SIGNAL
        'PropertiesChanged'
    

    **Objects:**
    
    * OBJECT_PATH
        '/org/mpris/MediaPlayer2'
    
    '''
    #interface
    MEDIA_PLAYER = 'org.mpris.MediaPlayer2'
    TRACK_LIST = 'org.mpris.MediaPlayer2.TrackList'
    PLAYER = 'org.mpris.MediaPlayer2.Player'
    PLAYLISTS = 'org.mpris.MediaPlayer2.Playlists'
    PROPERTIES = 'org.freedesktop.DBus.Properties'
    #signal
    SIGNAL = 'PropertiesChanged'
    #Object
    OBJECT_PATH = '/org/mpris/MediaPlayer2'



'''
From mprisV2.2 documentation

http://specifications.freedesktop.org/mpris-spec/latest/Track_List_Interface.html#Mapping:Metadata_Map
'''


class Metadata_Map(dict):
    '''
    A mapping from metadata attribute names to values.
    
    The mpris:trackid attribute must always be present. This contains a string that uniquely identifies the track within the scope of the playlist.
    
    If the length of the track is known, it should be provided in the metadata property with the 'mpris:length' key. The length must be given in microseconds, and be represented as a signed 64-bit integer.
    
    If there is an image associated with the track, a URL for it may be provided using the 'mpris:artUrl' key. For other metadata, fields defined by the Xesam ontology should be used, prefixed by 'xesam:'. See http://wiki.xmms2.xmms.se/wiki/MPRIS_Metadata for a list of common fields.
    
    Lists of strings should be passed using the array-of-string ('as') D-Bus type. Dates should be passed as strings using the ISO 8601 extended format (eg: 2007-04-29T14:35:51). If the timezone is known, RFC 3339's internet profile should be used (eg: 2007-04-29T14:35:51+02:00).
    
    * Attribute - s
        The name of the attribute; see http://wiki.xmms2.xmms.se/wiki/MPRIS_Metadata for guidelines on names to use.
    * Value - v
        The value of the attribute, in the most appropriate format.
    '''
    ART_URI = 'mpris:artUrl'
    TRACKID = 'mpris:trackid'
    LENGTH = 'mpris:length'
    ALBUM = 'xesam:album'
    ALBUM_ARTIST = 'xesam:albumArtist'
    ARTIST = 'xesam:artist'
    AS_TEXT = 'xesam:asText'
    AUDIO_BPM = 'xesam:audioBPM'
    AUTO_RATING = 'xesam:autoRating'
    COMMENT = 'xesam:comment'
    COMPOSER = 'xesam:composer'
    CONTENT_CREATED = 'xesam:contentCreated'
    DISC_NUMBER = 'xesam:discNumber'
    FIRST_USED = 'xesam:firstUsed'
    GENRE = 'xesam:genre'
    LAST_USED = 'xesam:lastUsed'
    LYRICIST = 'xesam:lyricist'
    TITLE = 'xesam:title'
    TRACK_NUMBER = 'xesam:trackNumber'
    URL = 'xesam:url'
    USE_COUNT = 'xesam:useCount'
    USER_RATING = 'xesam:userRating'

    def __init__(self, metadata):
        super(Metadata_Map, self).__init__(**metadata)


if __name__ == '__main__':
    mdm = Metadata_Map({Metadata_Map.ALBUM : 'Marcelo Nova Ao Vivo'})
    assert mdm[Metadata_Map.ALBUM] == 'Marcelo Nova Ao Vivo'



'''
From mprisV2.2 documentation

http://specifications.freedesktop.org/mpris-spec/latest/Player_Interface.html#Enum:Playback_Status
'''


class Playback_Status(str):
    '''
    A playback state.
    
    * Playing (Playing)
        A track is currently playing.
    * Paused (Paused)
        A track is currently paused.
    * Stopped (Stopped)
        There is no track currently playing.
        
    '''

    PLAYING = 'Playing'
    PAUSED = 'Paused'
    STOPPED = 'Stopped'
    VALUES = (PLAYING, PAUSED, STOPPED)
    
    def __init__(self, status):
        self._status = status
    
    def __int__(self):
        return Playback_Status.VALUES.index(self._status)
    
    def __str__(self):
        return self._status


if __name__ == '__main__':
    assert Playback_Status.PLAYING == 'Playing'
    assert Playback_Status.PAUSED == 'Paused'
    assert Playback_Status.STOPPED == 'Stopped'
    assert Playback_Status.STOPPED != 'Playing'



from .loop_status import Loop_Status
from .metadata_map import Metadata_Map
from .playback_rate import Playback_Rate
from .playback_status import Playback_Status
from .playlist import Playlist
from .playlist import Maybe_Playlist
from .playlist_id import Playlist_Id
from .playlist_ordering import Playlist_Ordering
from .time_in_us import Time_In_Us
from .uri import Uri
from .volume import Volume



from . import Loop_Status
from . import Metadata_Map
from . import Playback_Rate
from . import Playback_Status
from . import Playlist
from . import Maybe_Playlist
from . import Playlist_Id
from . import Playlist_Ordering
from . import Time_In_Us
from . import Uri
from . import Volume


if __name__ == '__main__':
    print(Loop_Status)
    print(Metadata_Map)
    print(Playback_Rate)
    print(Playback_Status)
    print(Playlist)
    print(Playlist_Id)
    print(Playlist_Ordering)
    print(Maybe_Playlist)
    print(Time_In_Us)
    print(Uri)
    print(Volume)


'''
This is not part of specification

Helper class to make it work as python lib
'''

I_PROP = 'org.freedesktop.DBus.Properties'
ARG_KEY = 'dbus_interface_info'
ATTR_KEY = '_dbus_interface_info'


class Decorator(object):
    def _update_me(self, target=None):
        if hasattr(target, "__doc__"):
            self.__doc__ = target.__doc__
        if hasattr(target, "__name__"):
            self.__name__ = target.__name__
        if hasattr(target, "__bases__"):
            self.__bases__ = target.__bases__



'''
From mprisV2.2 documentation

http://specifications.freedesktop.org/mpris-spec/2.2/Playlists_Interface.html#Simple-Type:Playlist_Id
'''


class Playlist_Id(str):
    '''
    Unique playlist identifier.
    '''

    def __init__(self, playlist_id):
        self._playlist_id = playlist_id
    
    def __str__(self):
        return self._status


if __name__ == '__main__':
    pid = Playlist_Id('id of a playlist')
    assert pid == 'id of a playlist'



'''
From mprisV2.2 documentation

http://specifications.freedesktop.org/mpris-spec/2.2/Track_List_Interface.html
'''

from .decorator import DbusAttr
from .decorator import DbusMethod
from .decorator import DbusSignal
from .decorator import DbusInterface
from .interfaces import Interfaces
from .types import Metadata_Map


class TrackList(Interfaces):
    '''
    Interface for TrackList (org.mpris.MediaPlayer2.TrackList)

    Provides access to a short list of tracks which were recently played or
    will be played shortly. This is intended to provide context to the 
    currently-playing track, rather than giving complete access to the media
    player's playlist.

    Example use cases are the list of tracks from the same album as the
    currently playing song or the Rhythmbox play queue.

    Each track in the tracklist has a unique identifier. The intention is that
    this uniquely identifies the track within the scope of the tracklist. In 
    particular, if a media item (a particular music file, say) occurs twice in
    the track list, each occurrence should have a different identifier. If a
    track is removed from the middle of the playlist, it should not affect the
    track ids of any other tracks in the tracklist.

    As a result, the traditional track identifiers of URLs and position in the
    playlist cannot be used. Any scheme which satisfies the uniqueness 
    requirements is valid, as clients should not make any assumptions about the
    value of the track id beyond the fact that it is a unique identifier.

    Note that the (memory and processing) burden of implementing the TrackList
    interface and maintaining unique track ids for the playlist can be
    mitigated by only exposing a subset of the playlist when it is very long
    (the 20 or so tracks around the currently playing track, for example). This
    is a recommended practice as the tracklist interface is not designed to
    enable browsing through a large list of tracks, but rather to provide
    clients with context about the currently playing track.
    '''
    PROPERTIES_TACKS = 'Tracks'
    PROPERTIES_CAN_EDIT_TRACKS = 'CanEditTracks'
    SIGNALS_TRACK_LIST_REPLACED = 'TrackListReplaced'
    SIGNALS_TRACK_ADDED = 'TrackAdded'
    SIGNALS_TRACK_REMOVED = 'TrackRemoved'
    SIGNALS_TRACK_METADATA_CHANGED = 'TrackMetadataChanged'
    SIGNALS_PROPERTIES_CHANGED = 'PropertiesChanged'

    @DbusInterface(Interfaces.TRACK_LIST, Interfaces.OBJECT_PATH)
    def __init__(self):
        '''Constructor'''
    
    @DbusMethod(produces=lambda map_list:\
                [Metadata_Map(metadata_map) for metadata_map in map_list])
    def GetTracksMetadata(self, TrackIds):
        '''
        **Parameters:**
        
        * TrackIds - ao (Track_Id_List)
            The list of track ids for which metadata is requested.
            
        **Returns**
        
        * Metadata - aa{sv} (Metadata_Map_List)
            Metadata of the set of tracks given as input.
            
        See the type documentation for more details.
        
        Gets all the metadata available for a set of tracks.
        
        Each set of metadata must have a 'mpris:trackid' entry at the very
        least, which contains a string that uniquely identifies this track
        within the scope of the tracklist.
        '''
        pass
    
    @DbusMethod()
    def AddTrack(self, Uri, AfterTrack='', SetAsCurrent=False):
        '''
        **Parameters:**
        
        * Uri - s (Uri)
            The uri of the item to add. Its uri scheme should be an element of
            the org.mpris.MediaPlayer2.SupportedUriSchemes property and the 
            mime-type should match one of the elements of the 
            org.mpris.MediaPlayer2.SupportedMimeTypes
        * AfterTrack - o (Track_Id)
            The identifier of the track after which the new item should be 
            inserted. The path /org/mpris/MediaPlayer2/TrackList/NoTrack 
            indicates that the track should be inserted at the start of the
            track list.
        * SetAsCurrent - b
            Whether the newly inserted track should be considered as the 
            current track. Setting this to trye has the same effect as calling
            GoTo afterwards.
        
        Adds a URI in the TrackList.
        
        If the CanEditTracks property is false, this has no effect.
        
        .. note::
            Clients should not assume that the track has been added at the time
            when this method returns. They should wait for a TrackAdded (or
            TrackListReplaced) signal.
        '''
        pass
    
    @DbusMethod
    def RemoveTrack(self, TrackId):
        '''
        
        **Parameters:**
        
        * TrackId - o (TrackId)
            Identifier of the track to be removed.
            /org/mpris/MediaPlayer2/TrackList/NoTrack is not a valid value for this argument. 
        
        Removes an item from the TrackList.
        
        If the track is not part of this tracklist, this has no effect.
        
        If the CanEditTracks property is false, this has no effect.
        
        .. note::
            Clients should not assume that the track has been removed at the
            time when this method returns. They should wait for a TrackRemoved
            (or TrackListReplaced) signal.
        '''
        pass
    
    @DbusMethod
    def GoTo(self, TrackId):
        '''
        **Parameters:**
        
        * TrackId - o (Track_Id)
            Identifier of the track to skip to.
            
            /org/mpris/MediaPlayer2/TrackList/NoTrack is not a valid value for this argument. 
        
        Skip to the specified TrackId.
        
        If the track is not part of this tracklist, this has no effect.
        
        If this object is not /org/mpris/MediaPlayer2, the current TrackList's
        tracks should be replaced with the contents of this TrackList, and the
        TrackListReplaced signal should be fired from /org/mpris/MediaPlayer2.
        '''
    
    @DbusSignal
    def TrackListReplaced(self, Tracks, CurrentTrack):
        '''
        **Parameters:**
        
        * Tracks - ao (Track_Id_List)
            The new content of the tracklist.
        * CurrentTrack - o (Track_Id)
            The identifier of the track to be considered as current.

            /org/mpris/MediaPlayer2/TrackList/NoTrack indicates that there is no current track.

            This should correspond to the mpris:trackid field of the Metadata property of 
            the org.mpris.MediaPlayer2.Player interface. 
    
        Indicates that the entire tracklist has been replaced.
        
        It is left up to the implementation to decide when a change to the 
        track list is invasive enough that this signal should be emitted 
        instead of a series of TrackAdded and TrackRemoved signals.
        '''
        pass
    
    @DbusSignal
    def TrackAdded(self, Metadata, AfterTrack=''):
        '''
        **Parameters:**
        
        * Metadata - a{sv} (Metadata_Map)
            The metadata of the newly added item.
            
            This must include a mpris:trackid entry.
            
            See the type documentation for more details.
        * AfterTrack - o (Track_Id)
            The identifier of the track after which the new track was inserted.
            The path /org/mpris/MediaPlayer2/TrackList/NoTrack indicates that
            the track was inserted at the start of the track list.


        Indicates that a track has been added to the track list.
        '''
        pass
    
    @DbusSignal
    def TrackRemoved(self, TrackId):
        '''
        **Parameters:**
        
        * TrackId - o (Track_Id)
            The identifier of the track being removed.
            
            /org/mpris/MediaPlayer2/TrackList/NoTrack is not a valid value for this argument. 

        Indicates that a track has been removed from the track list.
        '''
        pass
    
    @DbusSignal
    def TrackMetadataChanged(self, TrackId, Metadata):
        '''
        **Parameters:**
        
        * TrackId - o (Track_Id)
            The id of the track which metadata has changed.

            If the track id has changed, this will be the old value.

            /org/mpris/MediaPlayer2/TrackList/NoTrack is not a valid value for this argument.

        * Metadata - a{sv} (Metadata_Map)
            The new track metadata.    

            This must include a mpris:trackid entry.

            See the type documentation for more details.

        Indicates that the metadata of a track in the tracklist has changed.

        This may indicate that a track has been replaced, in which case the
        mpris:trackid metadata entry is different from the TrackId argument.
        '''
        pass
    
    @DbusAttr
    def Tracks(self):
        '''
        **Returns:**
        
        Read only
            When this property changes, the 
            org.freedesktop.DBus.Properties.PropertiesChanged signal is 
            emitted, but the new value is not sent.
        
        An array which contains the identifier of each track in the tracklist,
        in order.        

        The org.freedesktop.DBus.Properties.PropertiesChanged signal is emited
        every time this property changes, but the signal message does not
        contain the new value. Client implementations should rather rely on the
        TrackAdded, TrackRemoved and TrackListReplaced signals to keep their 
        representation of the tracklist up to date.
        '''
        pass
    
    @DbusAttr
    def CanEditTracks(self):
        '''
        **Returns:**
        
        Read only
            When this property changes, the 
            org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted
            with the new value.

        If false, calling AddTrack or RemoveTrack will have no effect, and may
        raise a NotSupported error.
        '''
        pass


if __name__ == '__main__':
    from .utils import get_players_uri, implements
    for uri in get_players_uri():
        if implements(uri, Interfaces.TRACK_LIST):
            # this is optional, maybe running player don't implement it
            # VLC does if you want try
            mp2 = TrackList(dbus_interface_info={'dbus_uri': uri})
            #print(dir(mp2))
            print('Tracks:', mp2.Tracks)
            print('CanEditTracks:', bool(mp2.CanEditTracks))
            break
    else:
        print('no player with playlist interface found')



from . import DbusAttr
from . import DbusInterface
from . import DbusMethod
from . import DbusSignal

if __name__ == '__main__':
    print(DbusAttr())
    print(DbusInterface())
    print(DbusMethod())
    print(DbusSignal())



'''
From mprisV2.2 documentation

http://specifications.freedesktop.org/mpris-spec/2.2/Playlists_Interface.html#Enum:Playlist_Ordering
'''


class Playlist_Ordering(str):
    '''
    Specifies the ordering of returned playlists.

    * Alphabetical (Alphabetical)
        Alphabetical ordering by name, ascending.
    * CreationDate (Created)
        Ordering by creation date, oldest first.
    * ModifiedDate (Modified)
        Ordering by last modified date, oldest first.
    * LastPlayDate (Played)
        Ordering by date of last playback, oldest first.
    * UserDefined (User)
        A user-defined ordering.
    '''

    ALPHABETICAL = 'Alphabetical'
    CREATION_DATE = 'CreationDate'
    MODIFIED_DATE = 'ModifiedDate'
    LAST_PLAY_DATE = 'LastPlayDate'
    USER_DEFINE = 'UserDefined'    
    VALUES = (ALPHABETICAL,CREATION_DATE,MODIFIED_DATE,LAST_PLAY_DATE,USER_DEFINE)
    
    def __init__(self, ordering):
        self._ordering = ordering

    def __str__(self):
        return self._ordering


if __name__ == '__main__':
    po = Playlist_Ordering('Alphabetical')
    assert po == 'Alphabetical'
    assert po == Playlist_Ordering.ALPHABETICAL
    assert po != Playlist_Ordering.CREATION_DATE



'''
From mprisV2.2 documentation

http://specifications.freedesktop.org/mpris-spec/latest/Player_Interface.html#Enum:Loop_Status
'''


class Loop_Status(str):
    '''
    A repeat / loop status
    
    * None (None)
        The playback will stop when there are no more tracks to play
    * Track (Track)
        The current track will start again from the begining once it has finished playing
    * Playlist (Playlist)
        The playback loops through a list of tracks
    '''
    NONE = 'None'
    TRACK = 'Track'
    PLAYLIST = 'Playlist'
    VALUES = (NONE, TRACK, PLAYLIST)
    def __init__(self, status):
        self._status = status
    
    def __int__(self):
        return Loop_Status.VALUES.index(self._status)
    
    def __str__(self):
        return self._status


if  __name__ == '__main__':
    assert Loop_Status.PLAYLIST != 'None'
    assert Loop_Status.PLAYLIST == 'Playlist'
    assert Loop_Status.TRACK == 'Track'
    assert Loop_Status.NONE == 'None'



'''
utils functions not defined in espec
'''

import re

from .interfaces import Interfaces
from .decorator.utils import get_session, get_uri, get_mainloop  # @UnusedImport
from .decorator.utils import list_interfaces as _list_interfaces
from .decorator.utils import implements as _implements

def get_players_uri(pattern=''):
    '''
        Return string of player bus name
        @param pattern=None: string RegEx that filter response
        @return: array string of players bus name
    '''
    return get_uri(Interfaces.MEDIA_PLAYER + '.*' + pattern)


def get_player_id_from_uri(uri):
    '''
        Returns player mpris2 id from uri
        @param uri: string mpris2 player dbus uri
        @return: string mrpis2 id
    '''
    mateched = re.match(Interfaces.MEDIA_PLAYER + '\.(.+)', uri or '')
    return mateched.groups()[0]\
        if mateched\
        else ''

def get_players_id(pattern=None):
    '''
        Return string of player mpris2 id
        @param pattern=None: string RegEx that filter response
        @return: array string of players bus name
    '''
    for item in get_uri(Interfaces.MEDIA_PLAYER + '.*' + pattern):
        yield get_player_id_from_uri(item)

def list_interfaces(uri, path=None, bus=None):
    return _list_interfaces(uri, path or Interfaces.OBJECT_PATH, bus)
        
def implements(uri, interface, path=Interfaces.OBJECT_PATH, bus=None):
    return _implements(uri, interface, path or Interfaces.OBJECT_PATH, bus)


if __name__ == '__main__':
    uris = get_players_uri()
    if not uris:
        print('No running players')
    for uri in uris:
        print(uri, ':')
        for interface in list_interfaces(uri):
            print('\t', interface)



from . import Interfaces
from . import MediaPlayer2
from . import Player
from . import Playlists
from . import TrackList


if __name__ == '__main__':
    print(Interfaces())
    print(MediaPlayer2())
    print(Player())
    print(Playlists())
    print(TrackList())



'''
This is not part of specification

Helper class to make it work as python lib
'''


from .attribute import DbusAttr
from .interface import DbusInterface
from .method import DbusMethod
from .signal import DbusSignal
from .utils import get_mainloop, get_uri, implements, \
list_all_interface, list_interfaces, list_paths



'''
From mprisV2.2 documentation

http://specifications.freedesktop.org/mpris-spec/latest/Media_Player.html
'''

from .decorator import DbusAttr
from .decorator import DbusInterface
from .decorator import DbusMethod
from .decorator import DbusSignal
from .interfaces import Interfaces


class MediaPlayer2(Interfaces):
    '''
    Interface for MediaPlayer2 (org.mpris.MediaPlayer2)
    '''
    PROPERTIES_CAN_QUIT = 'CanQuit'
    PROPERTIES_CAN_RAISE = 'Identity'
    PROPERTIES_HAS_TRACK_LIST = 'HasTrackList'
    PROPERTIES_IDENTITY = 'Identity'
    PROPERTIES_DESKTOP_ENTRY = 'DesktopEntry'
    PROPERTIES_SUPPORTED_URI_SCHEMES = 'SupportedUriSchemes'
    PROPERTIES_SUPPORTED_MINE_TYPES = 'SupportedMimeTypes'
    SIGNALS_PROPERTIES_CHANGED = 'PropertiesChanged'
    
    @DbusInterface(Interfaces.MEDIA_PLAYER, Interfaces.OBJECT_PATH)
    def __init__(self):
        '''Constructor'''
    
    @DbusMethod
    def Raise(self):
        '''
        Brings the media player's user interface to the front using any appropriate mechanism available.
        
        The media player may be unable to control how its user interface is displayed, or it may not have a graphical user interface at all. In this case, the Identity property is false and this method does nothing.
        '''
    
    @DbusMethod
    def Quit(self):
        '''
        Causes the media player to stop running.
        
        The media player may refuse to allow clients to shut it down. In this case, the CanQuit property is false and this method does nothing.
        
        ..note::
            Media players which can be D-Bus activated, or for which there is no sensibly easy way to terminate a running instance (via the main interface or a notification area icon for example) should allow clients to use this method. Otherwise, it should not be needed.
        
        If the media player does not have a UI, this should be implemented
        '''
    
    @DbusAttr
    def CanQuit(self):
        '''
        **Returns**
        
        Read only
            Inject attrs from decorator at new object then return obje
        
        When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.

        If false, calling Quit will have no effect, and may raise a NotSupported error. If true, calling Quit will cause the media application to attempt to quit (although it may still be prevented from quitting by the user, for example).
        '''

    @DbusAttr
    def Fullscreen(self):
        '''
        **Returns**
        
        Read Write
            Whether the media player is occupying the fullscreen.
            
        This property is optional. Clients should handle its absence gracefully.
        
        When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.
        
        This is typically used for videos. A value of true indicates that the media player is taking up the full screen.
        
        Media center software may well have this value fixed to true
        
        If CanSetFullscreen is true, clients may set this property to true to tell the media player to enter fullscreen mode, or to false to return to windowed mode.
        
        If CanSetFullscreen is false, then attempting to set this property should have no effect, and may raise an error. However, even if it is true, the media player may still be unable to fulfil the request, in which case attempting to set this property will have no effect (but should not raise an error).
        
        Added in 2.2.
        '''

    @DbusAttr    
    def CanSetFullscreen(self):
        '''
        **Returns**
        
        Read only
            If false, attempting to set Fullscreen will have no effect, and may raise an error. If true, attempting to set Fullscreen will not raise an error, and (if it is different from the current value) will cause the media player to attempt to enter or exit fullscreen mode.

        This property is optional. Clients should handle its absence gracefully.

        When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.
        
        Added in 2.2.
        
        ..note::
            Note that the media player may be unable to fulfil the request. In this case, the value will not change. If the media player knows in advance that it will not be able to fulfil the request, however, this property should be false.
        '''

    @DbusAttr
    def CanRaise(self):
        '''
        **Returns**
        
        Read only
            If false, calling Raise will have no effect, and may raise a NotSupported error. If true, calling Raise will cause the media application to attempt to bring its user interface to the front, although it may be prevented from doing so (by the window manager, for example).

        When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.
        '''
    
    @DbusAttr
    def HasTrackList(self):
        '''
        **Returns**
        
        Read only
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.
            
        Indicates whether the /org/mpris/MediaPlayer2 object implements the org.mpris.MediaPlayer2.TrackList interface.
        '''
    
    @DbusAttr
    def DesktopEntry(self):
        '''
        **Returns**
        
        Read only
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.

        The basename of an installed .desktop file which complies with the Desktop entry specification, with the '.desktop' extension stripped.

        Example: The desktop entry file is '/usr/share/applications/vlc.desktop', and this property contains 'vlc'

        This property is optional. Clients should handle its absence gracefully
        '''
    
    @DbusAttr
    def Identity(self):
        '''
        **Returns**
        
        Read only
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.

        If false, calling Raise will have no effect, and may raise a NotSupported error. If true, calling Raise will cause the media application to attempt to bring its user interface to the front, although it may be prevented from doing so (by the window manager, for example).
        '''
    
    @DbusAttr
    def SupportedUriSchemes(self):
        '''
        **Returns**
        
        Read only
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.

        The URI schemes supported by the media player.

        This can be viewed as protocols supported by the player in almost all cases. Almost every media player will include support for the 'file' scheme. Other common schemes are 'http' and 'rtsp'.

        Note that URI schemes should be lower-case.

        .. note::
            This is important for clients to know when using the editing capabilities of the Playlist interface, for example.
        '''
    
    @DbusAttr
    def SupportedMimeTypes(self):
        '''
        **Returns**
        
        Read only
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.

        The mime-types supported by the media player.

        Mime-types should be in the standard format (eg: audio/mpeg or application/ogg).

        .. note::
            This is important for clients to know when using the editing capabilities of the Playlist interface, for example.
        '''
    
    @DbusSignal(iface=Interfaces.PROPERTIES)
    def PropertiesChanged(self, *args, **kw):
        '''
        **Parameters:**
        
        * args - list
            unnamed parameters passed by dbus signal
        * kw - dict
            named parameters passed by dbus signal
            
        Every time that some property change, signal will be called
        '''


if __name__ == '__main__':
    from .utils import get_players_uri, implements, get_mainloop
    mainloop = get_mainloop()
    # to set signal handler
    # set_default_mainloop
    # need to be called before instance creation
    
    for uri in get_players_uri():
        if implements(uri, Interfaces.PLAYER):
            
            mp2 = MediaPlayer2(dbus_interface_info={
                'dbus_uri': uri
            })
            print(mp2.SupportedUriSchemes)
            
            def another_handler(self, *args, **kw): 
                print(args, '\n', kw)
        
            if mainloop:
                mp2.PropertiesChanged = another_handler
                mainloop.run()
            break
    else:
        print('no player with mediaplayer2 interface found')


'''
This is not part of specification

Helper class to make it work as python lib
'''

import dbus
from functools import wraps
from .base import Decorator, ARG_KEY, I_PROP, ATTR_KEY


class _DbusInfoProperty(object):
    def __init__(self, iface=None, path=None, 
                 uri=None, dbus_object=None, session=None, wrapped=None):
        self.iface = iface
        self.path = path
        self.uri = uri
        self.object = dbus_object
        self.session = session
        self.wrapped = wrapped
        self.interface = None
        self.properties = None
        
        if not self.object:
            bus = self.session = self.session or dbus.SessionBus()
            self.object = bus.get_object(self.uri, self.path)
        if not self.interface:
            self.interface = dbus.Interface(self.object,
                                            dbus_interface=self.iface)
        if not self.properties:
            self.properties = dbus.Interface(self.object, I_PROP)

    def reconnect(self, session=None):
        '''
        Required if you need update session/proxy object/interfaces
        '''
        
        session = session or self.session
        if session == self.session:
            self.session.close()
            session = self.session = dbus.SessionBus()
        self.object = session.get_object(self.uri, self.path)
        self.interface = dbus.Interface(self.object, dbus_interface=self.iface)
        self.properties = dbus.Interface(self.object, I_PROP)


class DbusInterface(Decorator):

    def __init__(self, iface=None, path=None, 
                 uri=None, dbus_object=None, session=None):
        self.iface = iface
        self.path = path
        self.uri = uri
        self.object = dbus_object
        self.session = session
        self.wrapped = None

    def __call__(self, meth):
        ''' Called when any decorated class is loaded'''
        self.wrapped = meth
        self._update_me(meth)
        
        @wraps(meth)
        def dbusWrapedInterface(*args, **kw):
            _args = kw.get(ARG_KEY, {})
            info_property = _DbusInfoProperty(
                iface=_args.get('dbus_iface', self.iface),
                path=_args.get('dbus_path', self.path),
                uri=_args.get('dbus_uri', self.uri),
                dbus_object =_args.get('dbus_object', self.object),
                session =_args.get('dbus_session', self.session),
                wrapped=self.wrapped
            )
            if ARG_KEY in kw: 
                del kw[ARG_KEY]
                
            return self.dbusWrapedInterface(info_property, *args, **kw)
        
        return dbusWrapedInterface
    
    def dbusWrapedInterface(self, info_property, *args, **kw):
        ''' Called when some decoreted class was called
        Inject attrs from decorator at new object then return object
        
        @param *args: list of args to call constructor
        @param **kw: dict of keywords, can redefine class default parameters
        @return: instance of decoreted class, with new attributes
        @see: mpris2.mediaplayer2 to see some examples
        '''
        #call decorated class constructor
        new_obj = self.wrapped(*args, **kw)
        if new_obj:
            setattr(new_obj, ATTR_KEY, info_property)
        elif len(args) > 0:
            setattr(args[0], ATTR_KEY, info_property)
        
        return new_obj


if __name__ == '__main__':
    # examples
    @DbusInterface('org.freedesktop.DBus', '/')
    class Example(object):
        pass


    d = Example(
        dbus_interface_info={
            'dbus_uri': 'org.freedesktop.DBus'})
    assert d._dbus_interface_info.iface == 'org.freedesktop.DBus'
    assert d._dbus_interface_info.path == '/'
    assert d._dbus_interface_info.uri == 'org.freedesktop.DBus'

    class ExempleToo(object):
        @DbusInterface('org.freedesktop.DBus', '/')
        def __init__(self):
            pass

    dd = ExempleToo(
        dbus_interface_info={
            'dbus_uri': 'org.freedesktop.DBus'})
        
    assert dd._dbus_interface_info.iface == 'org.freedesktop.DBus'
    assert dd._dbus_interface_info.path == '/'
    assert dd._dbus_interface_info.uri == 'org.freedesktop.DBus'



'''
From mprisV2.2 documentation

http://specifications.freedesktop.org/mpris-spec/latest/Player_Interface.html#Simple-Type:Volume
'''


class Volume(float):
    '''
    Audio volume level
    
    * 0.0 means mute.
    * 1.0 is a sensible maximum volume level (ex: 0dB).
    
    Note that the volume may be higher than 1.0, although generally clients should not attempt to set it above 1.0.
    '''

    MIN = 0.0
    MAX = 1.0
    RANGE = set([n/10.0 for n in range(11)])
    
    def __init__(self, volume=1.0):
        assert volume <= 1
        self._volume = float(volume)
    
    def __float__(self):
        return self._volume
    
    def __str__(self):
        return str(self._volume)


if __name__ == '__main__':
    assert Volume(1) == 1
    assert Volume(0.1) == 0.1
    assert Volume(1) == 1.0
    assert Volume(1.0) == 1
    assert Volume(0.1) != 1.2
    



from .base import Decorator, ATTR_KEY


class DbusSignal(Decorator):
    '''
    https://docs.python.org/2/howto/descriptor.html#properties
    '''

    def __init__(self, meth=None, iface=None):
        self.attr = meth
        self.handler = None
        self.iface = iface
        self._update_me(meth)

    def __call__(self, meth):
        self.attr = meth
        self._update_me(meth)
        return self
    
    def __get__(self, obj, objtype=None):
        if obj:
            return self.handler

        #static call
        return self 

    def __set__(self, obj, value):
        if obj:
            _dbus = getattr(obj, ATTR_KEY)
            interface = _dbus.interface
            def handle(*args, **kwds):
                h = self.handler
                h and h(*args, **kwds)

            if not self.handler:
                interface.connect_to_signal(self.attr.__name__, handle,
                                    dbus_interface=self.iface)
            self.handler = value
        else: #static call
            self.attr = value

    def __delete__(self, obj):
        self.handler = None


if __name__ == '__main__':
    from .interface import DbusInterface
    from .utils import get_mainloop
    mainloop = get_mainloop()
    print('mainloop', mainloop)
    
    @DbusInterface('org.mpris.MediaPlayer2.player',
                   '/org/mpris/MediaPlayer2')
    class Example(object):
        
        @DbusSignal
        def Seeked(self):
            pass


    d = Example(
        dbus_interface_info={
            'dbus_uri': 'org.mpris.MediaPlayer2.gmusicbrowser'})
    
    def handler(self, *args):
        print(args)
    
    d.Seeked = handler
    mainloop and mainloop.run()



'''
utils functions
'''

import dbus, re
import xml.etree.ElementTree as ET


I_INSPECT = 'org.freedesktop.DBus.Introspectable'

def _match_uri(name, pattern='.+'):
    '''
    Filter logic for get_players and get_player_uri
    @param name: string name to test
    @param pattern=None:  string RegEx to test
    @return: boolean
    '''
    return re.match(pattern, name)

def get_session():
    '''
        @return: dbus.SessionBus.get_session()
    '''
    return dbus.SessionBus.get_session()

def get_uri(pattern='.'):
    '''
    Return string of player bus name
    @param pattern=None: string RegEx that filter response
    @return: array string of players bus name
    '''
    for item in get_session().list_names():
        if _match_uri(item, pattern):
            yield item


def get_mainloop():
    '''
    @return: mainloop (with 'run' method)
    '''
    try:
        from dbus.mainloop.glib import DBusGMainLoop
        DBusGMainLoop(set_as_default=True)
        try:
            import gobject
            return gobject.MainLoop()
        except:
            pass
         
        try:
            import gi.repository.GLib
            return gi.repository.GLib.MainLoop()
        except:
            pass
    except:
        pass
    
    class PyQtMainLoop():
        def __init__(self, app):
            self.app = app
        
        def run(self):
            if hasattr(self.app, 'exec_'):
                self.app.exec_()
            if hasattr(self.app, 'exec'):
                method = getattr(self.app, 'exec')
                method(self.app)

    
    try:
        from dbus.mainloop.qt import DBusQtMainLoop
        DBusQtMainLoop(set_as_default=True)
        try:
            from PySide.QtGui import QApplication
            return PyQtMainLoop(QApplication([]))
        except:
            pass
    
        try:
            from PyQt5.QtWidgets import QApplication  
            return PyQtMainLoop(QApplication([]))
        except:
            pass
    
        try:
            from PyQt4 import Qt
            return PyQtMainLoop(Qt.QApplication([]))
        except:
            pass
    except:
        pass

    return None

def _introspect_uri(bus_name, path, bus):
    bus = bus or dbus.SessionBus.get_session()
    path = path.replace('//', '/') or '/'
    obj = bus.get_object(bus_name, path)
    xml = obj.Introspect(dbus_interface=I_INSPECT)
    return ET.fromstring(xml)

def list_paths(bus_name, path='/', bus=None):
    '''
    Return generator with all subpaths of uri
    @param bus_name: string dbus object name 
    @param path='/': string dbus object path
    @param bus=None: Conn dbus.Conn (commonly SessionBus or SystemBus)
    @return: generator with all paths that has interfaces
    '''
    nodes = _introspect_uri(bus_name, path, bus)
    if [iface for iface in nodes.findall('interface')]:
        yield path.replace('//', '/')

    for node in nodes.findall('node'):
        sub_path = path + '/'+ node.attrib['name']
        for path_name in list_paths(bus_name, sub_path, bus):
            yield path_name

def list_interfaces(bus_name, path, bus=None):
    '''
    Return generator with all interfaces of path 
    @param bus_name: string dbus object name 
    @param path: string dbus object path
    @param bus=None: Conn dbus.Conn (commonly SessionBus or SystemBus)
    @return: generator with all interfaces of path 
    '''
    nodes = _introspect_uri(bus_name, path, bus)    
    for interface in nodes.findall('interface'):
        yield interface.attrib['name']
        
def list_all_interface(bus_name, path='/', bus=None):
    '''
    Return generator with all interfaces of path and subpaths 
    @param bus_name: string dbus object name 
    @param path: string dbus object path
    @param bus=None: Conn dbus.Conn (commonly SessionBus or SystemBus)
    @return: generator with all interfaces of path and subpaths
    '''
    paths = list_paths(bus_name, path, bus)
    for sub_path in paths:
        for interface in list_interfaces(bus_name, sub_path, bus):
            yield sub_path, interface

def implements(uri, interface, path, bus=None):
    '''
    '''
    for iface in list_interfaces(uri, path, bus):
        if iface == interface:
            return True


if __name__ == '__main__':
    uri = None
    for uri in get_uri():
        if uri.count(':') == 0: 
            print(uri)
            
    if uri:
        for interface in list_all_interface(uri, '/'):
            print('\t', interface)
    else:
        print('Nothing running')



'''
From mprisV2.2 documentation

http://specifications.freedesktop.org/mpris-spec/latest/Player_Interface.html
'''

from .decorator import DbusAttr
from .decorator import DbusInterface
from .decorator import DbusMethod
from .decorator import DbusSignal
from .interfaces import Interfaces
from .types import Time_In_Us
from .types import Loop_Status
from .types import Playback_Status
from .types import Playback_Rate
from .types import Metadata_Map
from .types import Volume


class Player(Interfaces):
    '''
    This interface implements the methods for querying and providing basic control over what is currently playing.
    '''
    
    @DbusInterface(Interfaces.PLAYER, Interfaces.OBJECT_PATH)
    def __init__(self):
        '''Constructor'''
        
    
    @DbusMethod
    def Next(self):
        '''
        Skips to the next track in the tracklist.

        If there is no next track (and endless playback and track repeat are both off), stop playback.

        If playback is paused or stopped, it remains that way.

        If CanGoNext is false, attempting to call this method should have no effect.
        '''
        
    
    @DbusMethod 
    def Previous(self):
        '''
        Skips to the previous track in the tracklist.

        If there is no previous track (and endless playback and track repeat are both off), stop playback.

        If playback is paused or stopped, it remains that way.

        If CanGoPrevious is false, attempting to call this method should have no effect.
        '''
        
    
    @DbusMethod
    def Pause(self):
        '''
        Pauses playback.

        If playback is already paused, this has no effect.

        Calling Play after this should cause playback to start again from the same position.

        If CanPause is false, attempting to call this method should have no effect.
        '''
        
    
    @DbusMethod
    def PlayPause(self):
        '''
        Pauses playback.

        If playback is already paused, resumes playback.

        If playback is stopped, starts playback.

        If CanPause is false, attempting to call this method should have no effect and raise an error.
        '''
        
    
    @DbusMethod
    def Stop(self):
        '''
        Stops playback.

        If playback is already stopped, this has no effect.

        Calling Play after this should cause playback to start again from the beginning of the track.

        If CanControl is false, attempting to call this method should have no effect and raise an error.
        '''
        
    
    @DbusMethod
    def Play(self):
        '''
        Starts or resumes playback.

        If already playing, this has no effect.

        If there is no track to play, this has no effect.

        If CanPlay is false, attempting to call this method should have no effect.
        '''
        
    
    @DbusMethod
    def Seek(self, Offet):
        '''
        **Parameters:**

        * Offset - x (Time_In_Us)
            The number of microseconds to seek forward.
            
        Seeks forward in the current track by the specified number of microseconds.

        A negative value seeks back. If this would mean seeking back further than the start of the track, the position is set to 0.

        If the value passed in would mean seeking beyond the end of the track, acts like a call to Next.

        If the CanSeek property is false, this has no effect.
        '''
        
    
    @DbusMethod
    def SetPosition(self, TrackId, Position):
        '''
        **Parameters**
        
        * TrackId - o (Track_Id)
            The currently playing track's identifier.

            If this does not match the id of the currently-playing track, the call is ignored as 'stale'.
        * Position - x (Time_In_Us)
            Track position in microseconds.

            This must be between 0 and <track_length>.
            
        Sets the current track position in microseconds.

        If the Position argument is less than 0, do nothing.

        If the Position argument is greater than the track length, do nothing.

        If the CanSeek property is false, this has no effect.
        '''
        
    
    @DbusMethod
    def OpenUri(self, Uri):
        '''
        **Parameters:**
        
        * Uri - s (Uri)
            Uri of the track to load. Its uri scheme should be an element of the org.mpris.MediaPlayer2.SupportedUriSchemes property and the mime-type should match one of the elements of the org.mpris.MediaPlayer2.SupportedMimeTypes.
        
        Opens the Uri given as an argument

        If the playback is stopped, starts playing

        If the uri scheme or the mime-type of the uri to open is not supported, this method does nothing and may raise an error. In particular, if the list of available uri schemes is empty, this method may not be implemented.

        Clients should not assume that the Uri has been opened as soon as this method returns. They should wait until the mpris:trackid field in the Metadata property changes.

        If the media player implements the TrackList interface, then the opened track should be made part of the tracklist, the org.mpris.MediaPlayer2.TrackList.TrackAdded or org.mpris.MediaPlayer2.TrackList.TrackListReplaced signal should be fired, as well as the org.freedesktop.DBus.Properties.PropertiesChanged signal on the tracklist interface.
        '''
        
    
    @DbusSignal
    def Seeked(self, Position):
        '''
        **Parameters:**
        
        * Position - x (Time_In_Us)
            The new position, in microseconds.

        Indicates that the track position has changed in a way that is inconsistant with the current playing state.

        When this signal is not received, clients should assume that:

        * When playing, the position progresses according to the rate property.
        * When paused, it remains constant.

        This signal does not need to be emitted when playback starts or when the track changes, unless the track is starting at an unexpected position. An expected position would be the last known one when going from Paused to Playing, and 0 when going from Stopped to Playing.
        '''
        return Time_In_Us(Position)

    @DbusSignal(iface=Interfaces.PROPERTIES)
    def PropertiesChanged(self, *args, **kw):
        '''
        **Parameters**
        
        * args - list
            unnamed parameters passed by dbus signal
        * kw - dict
            named parameters passed by dbus signal
            
        Every time that some property change, signal will be called
        '''
        
    
    @DbusAttr(produces=Playback_Status)
    def PlaybackStatus(self):
        '''
        **Returns**
        
        Read only
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.
            
        The current playback status.
        
        May be 'Playing', 'Paused' or 'Stopped'.
        '''
        
    
    @DbusAttr(produces=Loop_Status)
    def LoopStatus(self):
        '''
        **Returns**
        
        Read/Write
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.

        The current loop / repeat status
        
        May be:
        
        * 'None' if the playback will stop when there are no more tracks to play
        * 'Track' if the current track will start again from the begining once it has finished playing
        * 'Playlist' if the playback loops through a list of tracks
        
        This property is optional, and clients should deal with NotSupported errors gracefully.
        
        If CanControl is false, attempting to set this property should have no effect and raise an error.
        '''
            

    @DbusAttr(produces=Playback_Rate)
    def Rate(self):
        '''
        **Returns**
        
        Read/Write
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.
        
        The current playback rate.
        
        The value must fall in the range described by MinimumRate and MaximumRate, and must not be 0.0. If playback is paused, the PlaybackStatus property should be used to indicate this. A value of 0.0 should not be set by the client. If it is, the media player should act as though Pause was called.
        
        If the media player has no ability to play at speeds other than the normal playback rate, this must still be implemented, and must return 1.0. The MinimumRate and MaximumRate properties must also be set to 1.0.
        
        Not all values may be accepted by the media player. It is left to media player implementations to decide how to deal with values they cannot use; they may either ignore them or pick a 'best fit' value. Clients are recommended to only use sensible fractions or multiples of 1 (eg: 0.5, 0.25, 1.5, 2.0, etc).
        '''
        

    @DbusAttr
    def Shuffle(self):
        '''
        **Returns**
        
        Read/Write
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.
        
        A value of false indicates that playback is progressing linearly through a playlist, while true means playback is progressing through a playlist in some other order.
        
        This property is optional, and clients should deal with NotSupported errors gracefully.
        
        If CanControl is false, attempting to set this property should have no effect and raise an error.
        '''
        

    @DbusAttr(produces=Metadata_Map)
    def Metadata(self):
        '''
        **Returns**
        
        Read only
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.
        
        The metadata of the current element.
        
        If there is a current track, this must have a 'mpris:trackid' entry at the very least, which contains a string that uniquely identifies this track.
        
        See the type documentation for more details.
        '''
        

    @DbusAttr(produces=Volume)
    def Volume(self):
        '''
        **Returns**
        
        Read/Write
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.
        
        The volume level.
        
        When setting, if a negative value is passed, the volume should be set to 0.0.
        
        If CanControl is false, attempting to set this property should have no effect and raise an error.
        '''
        

    @DbusAttr
    def Position(self):
        '''
        **Returns**
        
        Read only
            The org.freedesktop.DBus.Properties.PropertiesChanged signal is not emitted when this property changes.
        
        The current track position in microseconds, between 0 and the 'mpris:length' metadata entry (see Metadata).
        
        .. note::
            If the media player allows it, the current playback position can be changed either the SetPosition method or the Seek method on this interface. If this is not the case, the CanSeek property is false, and setting this property has no effect and can raise an error.
        
        If the playback progresses in a way that is inconstistant with the Rate property, the Seeked signal is emited.
        '''
        

    @DbusAttr
    def MinimumRate(self):
        '''
        **Returns**
        
        Read only
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.
        
        The minimum value which the Rate property can take. Clients should not attempt to set the Rate property below this value.
        
        Note that even if this value is 0.0 or negative, clients should not attempt to set the Rate property to 0.0.
        
        This value should always be 1.0 or less.
        '''
        

    @DbusAttr
    def MaximumRate(self):
        '''
        **Returns**
        
        Read only
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.
        
        The maximum value which the Rate property can take. Clients should not attempt to set the Rate property above this value.
        
        This value should always be 1.0 or greater.
        '''
        

    @DbusAttr
    def CanGoNext(self):
        '''
        **Returns**
        
        Read only
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.
        
        Whether the client can call the Next method on this interface and expect the current track to change.
        
        If CanControl is false, this property should also be false.
        '''
        
    
    @DbusAttr
    def CanGoPrevious(self):
        '''
        **Returns**
        
        Read only
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.
        
        Whether the client can call the Previous method on this interface and expect the current track to change.
        
        If CanControl is false, this property should also be false.
        '''
        
    
    @DbusAttr
    def CanPlay(self):
        '''
        **Returns**
        
        Read only
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.
        
        Whether playback can be started using Play or PlayPause.
        
        Note that this is related to whether there is a 'current track': the value should not depend on whether the track is currently paused or playing. In fact, if a track is currently playing CanControl is true), this should be true.
        
        If CanControl is false, this property should also be false.
        '''
        
        
    @DbusAttr
    def CanPause(self):
        '''
        **Returns**
        
        Read only
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.
        
        Whether playback can be paused using Pause or PlayPause.
        
        Note that this is an intrinsic property of the current track: its value should not depend on whether the track is currently paused or playing. In fact, if playback is currently paused (and CanControl is true), this should be true.
        
        If CanControl is false, this property should also be false.
        '''
    
    @DbusAttr
    def CanSeek(self):
        '''
        **Returns**
        
        Read only
            When this property changes, the org.freedesktop.DBus.Properties.PropertiesChanged signal is emitted with the new value.
        
        Whether the client can control the playback position using Seek and SetPosition. This may be different for different tracks.
        
        If CanControl is false, this property should also be false.
        '''
    
    @DbusAttr
    def CanControl(self):
        '''
        **Returns**
        
        Read only
            The org.freedesktop.DBus.Properties.PropertiesChanged signal is not emitted when this property changes.
        
        Whether the media player may be controlled over this interface.
        
        This property is not expected to change, as it describes an intrinsic capability of the implementation.
        
        If this is false, clients should assume that all properties on this interface are read-only (and will raise errors if writing to them is attempted); all methods are not implemented and all other properties starting with 'Can' are also false.
        '''

    
if __name__ == '__main__':
    from .utils import get_players_uri, implements
    for uri in get_players_uri():
        if implements(uri, Interfaces.PLAYER):
            mp2 = Player(dbus_interface_info={'dbus_uri': uri})
            print( mp2.LoopStatus )
            print( mp2.Shuffle )
            mp2.Shuffle = not mp2.Shuffle
            print( mp2.Shuffle )
            break
    else:
        print('no player with player interface found')